"Tests for parallel writers"

__copyright__ = "Copyright © 2018-2024 Lucas Frérot"
__license__ = "SPDX-License-Identifier: MIT"

import numpy as np
import pytest

from mpi4py import MPI
from vtk.util.numpy_support import vtk_to_numpy
from vtk import vtkXMLPRectilinearGridReader, vtkXMLPImageDataReader
from numpy import all

from conftest import get_vtk_data

from uvw.parallel import PRectilinearGrid, PImageData
from uvw import DataArray


def sequential_setup(uvw_type, vtk_type, filename, args, fixtures):
    (coords, r, e_r, field, order), compress, \
        vtk_format, comp_order, tmp_path = fixtures

    dim = r.ndim
    out_name = tmp_path / filename

    rect = uvw_type(out_name, *args(coords), dim * [0],
                    compression=compress,
                    byte_order=order)

    rect.addPointData(
        DataArray(
            r, range(dim), 'point', components_order=comp_order.param
        ), vtk_format=vtk_format) \
        .addCellData(DataArray(
            e_r, range(dim), 'cell', components_order=comp_order.param
        ), vtk_format=vtk_format) \
        .addFieldData(DataArray(
            field, [0], 'field', components_order=comp_order.param
        ), vtk_format=vtk_format).write()

    reader = vtk_type()
    vtk_r, vtk_e_r, vtk_f = get_vtk_data(reader, out_name)

    vtk_r = vtk_r.reshape(r.shape, order='F')
    vtk_e_r = vtk_e_r.reshape(e_r.shape, order='F') \
                     .transpose(comp_order.transp(dim))

    assert all(vtk_r == r)
    assert all(vtk_e_r == e_r)
    assert all(vtk_f == field)


@pytest.mark.mpi_skip
def test_prectilinear_grid(field_data,
                           compression_fixture,
                           format_fixture,
                           ordering_fixture,
                           tmp_path):
    sequential_setup(PRectilinearGrid,
                     vtkXMLPRectilinearGridReader,
                     'test_prectilinear_grid.pvtr',
                     lambda coords: [coords],
                     (field_data, compression_fixture.param,
                      format_fixture.param, ordering_fixture, tmp_path))


@pytest.mark.mpi_skip
def test_pimagedata(field_data,
                    compression_fixture,
                    format_fixture,
                    ordering_fixture,
                    tmp_path):
    sequential_setup(PImageData,
                     vtkXMLPImageDataReader,
                     'test_pimage_data.pvtr',
                     lambda coords: [[(x.min(), x.max()) for x in coords],
                                     [x.size for x in coords]],
                     (field_data, compression_fixture.param,
                      format_fixture.param, ordering_fixture, tmp_path))


def parallel_setup(uvw_type, vtk_type, filename, args, fixtures):
    compress, vtk_format, tmp_path = fixtures

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()

    N = 3

    bounds = [
        (0, 1),
        (1, 2)
    ]

    offsets = [
        [0, 0, 0],
        [2*N, 0, 0],
    ]

    x = np.linspace(*bounds[rank], N*2)
    y = np.linspace(0, 1, N+2)
    z = np.linspace(0, 1, N)

    tmp_path = comm.bcast(tmp_path, root=0)
    out_name = tmp_path / filename
    print(out_name)

    xx, yy, zz = np.meshgrid(x, y, z, indexing='ij', sparse=True)
    r = np.sqrt(xx**2 + yy**2 + zz**2)

    rect = uvw_type(out_name, *args((x, y, z)), offsets[rank],
                    compression=compress)

    rect.addPointData(DataArray(r, range(3), 'R'), vtk_format=vtk_format)
    rect.write()

    if rank == 0:
        reader = vtk_type()
        reader.SetFileName(str(out_name))
        reader.Update()
        output = reader.GetOutput()
        vtk_r = vtk_to_numpy(output.GetPointData().GetArray('R'))
    else:
        vtk_r = None

    vtk_r = comm.bcast(vtk_r, root=0)
    vtk_r = vtk_r.reshape([4*N-1, N+2, N], order='F')

    i, j, k = [int(i) for i in offsets[rank]]

    # Adjusting for overlap
    if offsets[rank][0] != 0:
        i -= 1

    sub_vtk = vtk_r[i:i+x.size, j:j+y.size, k:k+z.size]
    assert np.all(sub_vtk == r)


@pytest.mark.mpi(min_size=2)
def test_prectilinear_grid_mpi(compression_fixture,
                               format_fixture,
                               tmp_path):
    parallel_setup(PRectilinearGrid, vtkXMLPRectilinearGridReader,
                   'test_prectilinear_grid_mpi.pvtr',
                   lambda coords: [coords],
                   (compression_fixture.param, format_fixture.param, tmp_path))


@pytest.mark.mpi(min_size=2)
def test_pimagedata_mpi(compression_fixture,
                        format_fixture,
                        tmp_path):
    parallel_setup(PImageData, vtkXMLPImageDataReader,
                   'test_pimage_data_mpi.pvti',
                   lambda coords: [[(x.min(), x.max()) for x in coords],
                                   [x.size for x in coords]],
                   (compression_fixture.param, format_fixture.param, tmp_path))


if __name__ == '__main__':
    from pathlib import Path

    class fixture_stub:
        pass

    compression = fixture_stub()
    format = fixture_stub()

    compression.param = None
    format.param = 'ascii'

    test_prectilinear_grid_mpi(compression, format, Path('.'))



__copyright__ = "Copyright © 2018-2024 Lucas Frérot"
__license__ = "SPDX-License-Identifier: MIT"

import io
import numpy as np

from numpy import all, min, max
from vtk import (
    vtkXMLRectilinearGridReader,
    vtkXMLImageDataReader,
    vtkXMLStructuredGridReader,
    vtkXMLUnstructuredGridReader,
)
from vtk.util.numpy_support import vtk_to_numpy
from conftest import get_vtk_data

from uvw import (
    ImageData,
    RectilinearGrid,
    StructuredGrid,
    UnstructuredGrid,
    ParaViewData,
    DataArray,
)

from uvw.unstructured import CellType


def test_rectilinear_grid(field_data,
                          compression_fixture,
                          format_fixture,
                          ordering_fixture):
    coords, r, e_r, field, order = field_data
    dim = r.ndim
    f = io.StringIO()

    compress = compression_fixture.param
    format = format_fixture.param
    rect = RectilinearGrid(f, coords, compression=compress, byte_order=order)
    rect.addPointData(
        DataArray(r, range(dim), 'point', ordering_fixture.param),
        vtk_format=format
    ).addCellData(
        DataArray(e_r, range(dim), 'cell', ordering_fixture.param),
        vtk_format=format
    ).addFieldData(
        DataArray(field, [0], 'field', ordering_fixture.param),
        vtk_format=format
    )

    rect.write()

    reader = vtkXMLRectilinearGridReader()

    # Testing the xml pretty print output as well
    pretty_sstream = io.StringIO(str(rect.writer))

    for ss in [f, pretty_sstream]:
        vtk_r, vtk_e_r, vtk_f = get_vtk_data(reader, ss)

        vtk_r = vtk_r.reshape(r.shape, order='F')
        vtk_e_r = vtk_e_r.reshape(e_r.shape, order='F') \
                         .transpose(ordering_fixture.transp(dim))

        assert all(vtk_r == r)
        assert all(vtk_e_r == e_r)
        assert all(vtk_f == field)


def test_image_data(field_data,
                    compression_fixture,
                    format_fixture,
                    ordering_fixture):
    coords, r, e_r, field, order = field_data
    dim = r.ndim
    f = io.StringIO()

    compress = compression_fixture.param
    format = format_fixture.param
    with ImageData(
            f,
            [(min(x), max(x)) for x in coords],
            [x.size for x in coords],
            compression=compress,
            byte_order=order) as fh:
        fh.addPointData(
            DataArray(r, range(dim), 'point', ordering_fixture.param),
            vtk_format=format
        ).addCellData(
            DataArray(e_r, range(dim), 'cell', ordering_fixture.param),
            vtk_format=format
        ).addFieldData(
            DataArray(field, [0], 'field', ordering_fixture.param),
            vtk_format=format
        )

    reader = vtkXMLImageDataReader()
    vtk_r, vtk_e_r, vtk_f = get_vtk_data(reader, f)

    vtk_r = vtk_r.reshape(r.shape, order='F')
    vtk_e_r = vtk_e_r.reshape(e_r.shape, order='F') \
                     .transpose(ordering_fixture.transp(dim))

    assert all(vtk_r == r)
    assert all(vtk_e_r == e_r)
    assert all(vtk_f == field)


def test_structured_grid(compression_fixture, format_fixture):
    f = io.StringIO()

    N = 5

    r = np.linspace(0, 1, N)
    theta = np.linspace(0, 2*np.pi, 5*N)

    theta, r = np.meshgrid(theta, r, indexing='ij')

    x = r*np.cos(theta)
    y = r*np.sin(theta)

    points = np.vstack([x.ravel(), y.ravel()]).T

    compress = compression_fixture.param
    format = format_fixture.param
    grid = StructuredGrid(f, points, (N, 5*N), compression=compress)

    data = np.exp(-4*r**2)

    grid.addPointData(DataArray(data, reversed(range(2)), 'data'),
                      vtk_format=format)
    grid.write()

    reader = vtkXMLStructuredGridReader()
    reader.SetReadFromInputString(True)
    reader.SetInputString(f.getvalue())
    reader.Update()

    output = reader.GetOutput()
    vtk_data = vtk_to_numpy(output.GetPointData().GetArray('data'))
    vtk_data = vtk_data.reshape(data.shape, order='C')

    assert all(vtk_data == data)


def test_unstructured_grid(compression_fixture, format_fixture):
    f = io.StringIO()
    compress = compression_fixture.param
    format = format_fixture.param

    nodes = np.array([
        [0, 0, 0],
        [1, 0, 0],
        [0, 1, 0],
    ])

    point_data = np.array([[0, 1], [1, 2], [2, 3]])
    cell_data = np.array([1, 2, 3, 4])

    connectivity = {
        CellType.TRIANGLE: np.array([range(3)], dtype=np.int32),
        6: np.array([[0, 1, 2]], dtype=np.int32),  # Testing true VTK type id
        CellType.POLY_LINE: np.array([  # Testing variable length cell type
            [0, 1],
            [1, 2, 0],
        ], dtype=object),
    }

    grid = UnstructuredGrid(f, nodes, connectivity, compression=compress)
    grid.addPointData(DataArray(point_data, [0], 'point'), vtk_format=format)
    grid.addCellData(DataArray(cell_data, [0], 'cell'), vtk_format=format)
    grid.write()

    reader = vtkXMLUnstructuredGridReader()
    reader.SetReadFromInputString(True)
    reader.SetInputString(f.getvalue())
    reader.Update()

    vtk_pdata = vtk_to_numpy(
        reader.GetOutput().GetPointData().GetArray('point'))
    vtk_cdata = vtk_to_numpy(reader.GetOutput().GetCellData().GetArray('cell'))
    vtk_pdata.reshape(point_data.shape, order='C')
    vtk_cdata.reshape(cell_data.shape, order='C')
    assert all(vtk_pdata == point_data)
    assert all(vtk_cdata == cell_data)


def test_paraview_data(tmp_path):
    """
    NB: This is just testing writing. Since PVD is a ParaView related extension,
    it cannot be tested with vanilla VTK
    """
    x = np.linspace(0, 1, 10)
    y = x.copy()

    grid = RectilinearGrid(tmp_path / 'grid.vtr', [x, y])
    grid.write()

    group = ParaViewData(tmp_path / 'grid.pvd')
    group.addFile(grid)
    group.write()



__copyright__ = "Copyright © 2018-2024 Lucas Frérot"
__license__ = "SPDX-License-Identifier: MIT"

import pytest
import numpy as np

from uvw import (
    DataArray,
    RectilinearGrid,
    StructuredGrid,
    UnstructuredGrid,
)

from uvw.vtk_files import WriteManager
from uvw.unstructured import check_connectivity, CellType
from uvw.parallel import PRectilinearGrid


def test_malformed_attributes():
    x = np.array([0, 1])

    rect = RectilinearGrid('', x)

    with pytest.raises(ValueError):
        rect.piece.register('', [])


def test_unsupported_format():
    x = np.array([0, 1])

    rect = RectilinearGrid('', x)
    with pytest.raises(ValueError):
        rect.addPointData(DataArray(x, [0]), '#dummy')


def test_invalid_compression():
    x = np.array([0, 1])
    with pytest.raises(ValueError):
        RectilinearGrid('', x, compression=100)


def test_invalid_file():
    x = np.array([0, 1])
    grid = RectilinearGrid('', x)
    with pytest.raises(TypeError):
        grid.writer.write(1)
    with pytest.raises(TypeError):
        PRectilinearGrid(1, x, None)


def test_array_dimensions():
    x = np.array([0, 1])
    with pytest.raises(ValueError):
        DataArray(x, range(2))

    with pytest.raises(ValueError):
        DataArray(x, range(1), components_order=2)


def test_rectilinear_fails():
    x = np.array([[0, 1], [2, 3]])
    with pytest.raises(ValueError):
        RectilinearGrid('', x)

    x = np.array([0, 1])
    with pytest.raises(ValueError):
        RectilinearGrid('', x, [])


def test_structured_fails():
    x = np.array([1, 2])
    with pytest.raises(ValueError):
        StructuredGrid('', x, (1, 2))

def test_context_manger():
    x = np.array([1, 2])
    with RectilinearGrid('', x):
        raise Exception('Yo')

    with pytest.raises(NotImplementedError):
        with WriteManager():
            pass


def test_check_connectivity():
    connectivity = {CellType.TRIANGLE: [[0, 1, 2]]}
    with pytest.raises(TypeError):
        check_connectivity(connectivity)

    connectivity = {'triangle': np.array([[0, 1, 2]])}
    with pytest.raises(ValueError):
        check_connectivity(connectivity)

    connectivity = {CellType.TRIANGLE: np.array([[0., 1, 2]])}
    with pytest.raises(TypeError):
        check_connectivity(connectivity)

    connectivity = {5: np.array([[0, 1, 2, 3]])}
    assert not check_connectivity(connectivity)


def test_check_unstructured():
    nodes = np.zeros([1])
    with pytest.raises(ValueError):
        UnstructuredGrid('', nodes, {})

    nodes = np.zeros([1, 2])
    connectivity = {CellType.TRIANGLE: np.array([[0, 1, 2, 3]], dtype=np.int32)}
    with pytest.raises(ValueError):
        UnstructuredGrid('', nodes, connectivity)


def test_check_array_type_error():
    array = np.array([0, 1, 2], dtype=np.complex128)
    with pytest.raises(TypeError):
        DataArray(array, [0], '')


def test_invalid_endian():
    with pytest.raises(ValueError):
        x = np.array([0, 1])
        RectilinearGrid('', x, byte_order='WhatEndian')
